import os
import numpy as np
import math
import six
import iris
import iris.cube
from iris.cube import CubeList
import iris.analysis
import iris.util
import iris.coords
import warnings
import numba
import datetime
from numba import vectorize, guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

MESH_FILE = "/esarchive/autosubmit/con_files/mesh_mask_nemo.Ec3.2_O1L75.nc"
REGIONS_FILE = "/esarchive/autosubmit/con_files/mask.regions.Ec3.2_O1L75.nc"
VO_FILE = '/esarchive/exp/ecearth/a16l/cmorfiles/CMIP/EC-Earth-Consortium' \
          '/EC-Earth3-LR/historical/r1i1p1f1/Omon/vo/gn/v20180711' \
          '/vo_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_196101-196112.nc'
RES = '/esarchive/exp/ecearth/a16l/cmorfiles/CMIP/EC-Earth-Consortium' \
      '/EC-Earth3-LR/historical/r1i1p1f1/Omon/vsftmyz/gn/v20180711' \
      '/vsftmyz_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_196101-196112.nc'


def main():
    e3v, e1v = load_mesh()
    basin_name, basins = load_masks()
    area = {}
    for basin in basins:
        area[basin] = compute_area_basin(e1v, e3v, basins[basin])
    del e3v, e1v
    vo = load_vo()
    start = datetime.datetime.now()
    vsftmyz = compute_vsftmyz_cpu(vo, area)
    print('np {0}'.format(datetime.datetime.now()-start))
    cube = iris.cube.Cube(vsftmyz)
    iris.save(cube, '/esarchive/scratch/sloosvel/numba_outputs'
              '/vsftmyz_global_numpy.nc', zlib=True)


def compute_vsftmyz_cpu(vo, area):
    vsftmyx = []
    for basin in area:
        vsftmyz_basin = np.sum(multiply_vo_basin(vo, area[basin]), axis=3)
        compute_moc(vsftmyz_basin, vsftmyz_basin)
        vsftmyx.append(vsftmyz_basin)
    return vsftmyz_basin


def compute_vsftmyz_gpu(vo, area):
    levels = vo.shape[1]
    times = vo.shape[0]
    lats = vo.shape[2]
    lons = vo.shape[3]

    block = (128, 1, 1)
    grid_size = (lats // block[0]) + 1
    grid = (grid_size, levels, times)
    gpu_vo = cuda.to_device(vo.astype(np.float32))
    for basin in area:
        gpu_area = cuda.to_device(area[basin].astype(np.float32))
        gpu_moc = cuda.device_array((times, levels, lats), dtype=np.float32)
        horizontal_integral[grid, block](gpu_vo, gpu_area, gpu_moc)
        compute_moc_gpu(gpu_moc, gpu_moc)
        vsftmyz_basin = gpu_moc.copy_to_host()

    return vsftmyz_basin


def load_vo():
    vo = iris.load_cube(VO_FILE, 'sea_water_y_velocity')
    vo_data = np.ma.filled(vo.data, 0.0).astype(np.float32)
    return vo_data


def load_mesh():
    e3v = iris.load_cube(MESH_FILE, 'e3v_0').data.astype(np.float32)
    e3v = np.squeeze(e3v)
    e1v = iris.load_cube(MESH_FILE, 'e1v').data.astype(np.float32)
    e1v = np.squeeze(e1v)
    return e3v, e1v


def load_masks():
    basins = {}
    cubes = iris.load_cube(MESH_FILE, 'vmask')
    basin_name = []
    basin_name.append('a')
    cubes.data[..., 0] = 0.0
    cubes.data[..., -1] = 0.0
    basins['a'] = cubes.data.astype(np.float32)
    return basin_name, basins


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def compute_area_basin(e1v, e3v, basin):
    return - e1v * e3v * basin / 1e6


@vectorize(['float32(float32, float32)'], target='cpu')
def multiply_vo_basin(vo, basin):
    return vo*basin


@guvectorize([(float32[:, :, :], float32[:, :, :])], '(t, l, j)->(t, l, j)',
             target='cpu')
def compute_moc(moc, out):
    for lev in reversed(range(moc.shape[1]-1)):
        moc[:, lev, :] = moc[:, lev, :] + moc[:, lev+1, :]


@guvectorize([(float32[:, :, :], float32[:, :, :])], '(t, l, j)->(t, l, j)',
             target='cuda')
def compute_moc_gpu(moc, out):
    for lev in reversed(range(moc.shape[1]-1)):
        moc[:, lev, :] = moc[:, lev, :] + moc[:, lev+1, :]


@cuda.jit()
def horizontal_integral(vo, area, moc):
    j = cuda.threadIdx.x + cuda.blockIdx.x * cuda.blockDim.x
    lev = cuda.blockIdx.y
    t = cuda.blockIdx.z
    # i, j, t = grid(3)
    moc[t, lev, j] = 0.0
    temp = 0.0
    if(j >= vo.shape[2]):
        return
    for i in range(vo.shape[3]):
        temp += vo[t, lev, j, i] * area[0, lev, j, i]
    moc[t, lev, j] = temp


if __name__ == '__main__':
    main()
